Library IEEE;
USE IEEE.std_logic_1164.all;
USE IEEE.std_logic_unsigned.all;

USE work.pkg_081006.all;

----------------------------------------------------
-- This module computes the average of 4 sample, for
-- 4 channels and performs baseline subtraction.

entity Average_Channels_v2 is
port(
		clk				: in std_logic;
		nreset			: in std_logic;
		Run_DAQ			: in std_logic; -- resets the averaging process
		
		ADC_STRB		: in std_logic;
		CH1_D			: in std_logic_vector(15 downto 0);                             
		CH2_D			: in std_logic_vector(15 downto 0);                             
		CH3_D			: in std_logic_vector(15 downto 0);                             
		CH4_D			: in std_logic_vector(15 downto 0); 
		
		RD_Avg_Reg		: in std_logic_vector( 4 downto 1);
		Avg_Length		: in std_logic_vector(15 downto 0);

		AVG_TO_VME		: out std_logic_vector(15 downto 0);
		AVG1_OUT	 	: out std_logic_vector(15 downto 0);                            
		AVG2_OUT	 	: out std_logic_vector(15 downto 0);                            
		AVG3_OUT	 	: out std_logic_vector(15 downto 0);                            
		AVG4_OUT	 	: out std_logic_vector(15 downto 0);
		AVG_OUT_STRB 	: out std_logic;
		REMAINDER		: out std_logic_vector(15 downto 0)                        		
);
end Average_Channels_v2;

--########################################################
--				A R C H I T E C T U R E
--########################################################
architecture main of Average_Channels_v2 is
-- Synchronized Inputs --------------------------------
	signal rst_avg		: std_logic;
	signal get_adc		: std_logic;
	
	signal rd_sync		: std_logic_vector(4 downto 1);
	signal RD_Avg_Strb	: std_logic_vector(4 downto 1);

-- Control Signals ------------------------------------
	signal chan_cnt		: std_logic_vector(2 downto 0);
	signal inc_chan_cnt	: std_logic;
	signal rst_chan_cnt	: std_logic;
	
	signal sum_cnt		: std_logic_vector(15 downto 0);
	signal inc_sum_cnt	: std_logic;
	signal rst_sum_cnt	: std_logic;

	signal divide_pause	: std_logic_vector(3 downto 0);
	signal inc_pause	: std_logic;
	signal rst_pause	: std_logic;
	
	signal clr_sum_reg	: std_logic; -- clear the accumulation
	signal get_sum		: std_logic; -- latch output of adder
	signal get_avg		: std_logic; -- latch final value	
	signal all_avgs_rdy	: std_logic; -- latch all averages

-- Buffers ----------------------------------------------
	signal CH1_IN		: std_logic_vector(15 downto 0);
	signal CH2_IN		: std_logic_vector(15 downto 0);
	signal CH3_IN		: std_logic_vector(15 downto 0);
	signal CH4_IN		: std_logic_vector(15 downto 0);
			 
	signal CH1_SUM		: std_logic_vector(31 downto 0);
	signal CH2_SUM		: std_logic_vector(31 downto 0);
	signal CH3_SUM		: std_logic_vector(31 downto 0);
	signal CH4_SUM		: std_logic_vector(31 downto 0);
	
	signal AVG1_BUF		: std_logic_vector(15 downto 0);
	signal AVG2_BUF		: std_logic_vector(15 downto 0);
	signal AVG3_BUF		: std_logic_vector(15 downto 0);
	signal AVG4_BUF		: std_logic_vector(15 downto 0);
	
	signal NEW_VALUE	: std_logic_vector(31 downto 0);
	signal CURRENT_SUM	: std_logic_vector(31 downto 0);
	
-- Altera Math Functions ------------------------------
component unsigned_add32
	PORT
	(
		dataa		: IN STD_LOGIC_VECTOR (31 DOWNTO 0);
		datab		: IN STD_LOGIC_VECTOR (31 DOWNTO 0);
		result		: OUT STD_LOGIC_VECTOR (31 DOWNTO 0)
	);
end component;

	signal Sum_Result	: std_logic_vector(31 downto 0);

component unsigned_divide
	PORT
	(
		denom			: IN STD_LOGIC_VECTOR (15 DOWNTO 0);
		numer			: IN STD_LOGIC_VECTOR (31 DOWNTO 0);
		quotient		: OUT STD_LOGIC_VECTOR (31 DOWNTO 0);
		remain			: OUT STD_LOGIC_VECTOR (15 DOWNTO 0)
	);
end component;

	signal Average_Result	: std_logic_vector(31 downto 0);

-- SETUP_STATE_MACHINE ---------------------------------------
type STATE_TYPE is
(init, clr_sum, wait1, do_sum, ltch_sum, end_sample, do_div, wr_avg, avg_rdy);
signal CS, NS : STATE_TYPE;

--========================================================
--						B E G I N
--========================================================
begin

SYNC_RESET: led 
	port map(sig_in=>Run_DAQ, clk=>clk,	clrn=>nreset, pls_out=>rst_avg); 

SYNC_ADC: led 
	port map(sig_in=>ADC_STRB, clk=>clk,	clrn=>nreset, pls_out=>get_adc); 

----------------------------------------------------------
INPUT_BUFFERS: 
process(clk, nreset, ADC_STRB)
begin
	if(nreset='0')			then
		CH1_IN		<= (others => '0');	
		CH2_IN		<= (others => '0');			
		CH3_IN		<= (others => '0');			
		CH4_IN		<= (others => '0');			

	elsif(rising_edge(clk))	then
		if(ADC_STRB = '1') then
			CH1_IN		<= CH1_D;	
			CH2_IN		<= CH2_D;			
			CH3_IN		<= CH3_D;			
			CH4_IN		<= CH4_D;
		end if;			   
		 		
	end if;
end process;

-----------------------------------------------------------
CHANNEL_UPDATE_COUNT:
process(clk, nreset, rst_chan_cnt, inc_chan_cnt)
begin
	if  (nreset = '0')			then chan_cnt <= (others =>'0');
	elsif(rising_edge(clk))		then
		if   (rst_chan_cnt = '1') then chan_cnt <= (others =>'0');
		elsif(inc_chan_cnt = '1') then chan_cnt <= chan_cnt + 1;
		else						   chan_cnt <= chan_cnt;
		end if;			   
	end if;
end process;

----------------------------------------------------------
SAMPLES_AVERAGED_COUNT:
process(clk, nreset, rst_sum_cnt, inc_sum_cnt)
begin
	if  (nreset = '0')			 then sum_cnt <= (others =>'0');
	elsif(rising_edge(clk))		 then
		if   (rst_sum_cnt = '1') then sum_cnt <= (others =>'0');
		elsif(inc_sum_cnt = '1') then sum_cnt <= sum_cnt + 1;
		else						  sum_cnt <= sum_cnt;
		end if;			   
	end if;
end process;

----------------------------------------------------------
PAUSE_FOR_DIVISION:
process(clk, nreset, rst_pause, inc_pause)
begin
	if  (nreset = '0')		then divide_pause <= (others =>'0');
	elsif(rising_edge(clk))	then
		if   (rst_pause = '1') then divide_pause <= (others =>'0');
		elsif(inc_pause = '1') then divide_pause <= divide_pause + 1;
		else						divide_pause <= divide_pause;
		end if;			   
	end if;
end process;

----------------------------------------------------------
--{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{
--                S T A T E   M A C H I N E
--}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}
SM_REG: process(clk, nreset, rst_avg)
begin
	if(nreset='0' OR rst_avg = '1')	then CS <= init;
	elsif(rising_edge(clk))			then CS <= NS;
	end if;
end process;

SM_NEXT_STATE: 
process(CS, chan_cnt, get_adc, sum_cnt, divide_pause, Avg_Length)
begin
	-- defaults ----------------------
	clr_sum_reg		<= '0';
	get_sum			<= '0';
	get_avg			<= '0';
	all_avgs_rdy	<= '0';
	
	inc_chan_cnt	<= '0'; -- count to 4 channels processed
	rst_chan_cnt	<= '0';
	
	inc_sum_cnt		<= '0'; -- count to Average Length
	rst_sum_cnt		<= '0';
	
	inc_pause		<= '0'; -- timer to let division settle
	rst_pause		<= '0';
	
	case CS is
		when init =>						NS <= clr_sum;				
			rst_sum_cnt		<= '1';
			rst_chan_cnt	<= '1';

		when clr_sum =>						NS <= wait1;	
			clr_sum_reg		<= '1';

		when wait1 =>			
			if   (get_adc = '1')		then NS <= do_sum;
			elsif(sum_cnt = Avg_Length)	then NS <= do_div;
			else                   	     	 NS <= wait1;
			end if;
		
		when do_sum   =>					NS <= ltch_sum;	

		when ltch_sum =>				
			get_sum		<= '1';
			
			if(chan_cnt < "011")then 		NS <= do_sum;
			else                   	    	NS <= end_sample;
			end if;
			
			inc_chan_cnt<= '1'; -- count increments when leaving this state
		
		when end_sample =>					NS <= wait1;
			inc_sum_cnt	<= '1';
			rst_chan_cnt<= '1';
			
		-----------------------------------------------
		when do_div =>					
			rst_sum_cnt	<= '1';
			inc_pause	<= '1';
			
			if(divide_pause < "0110") then 	NS <= do_div;
			else                   	    	NS <= wr_avg;
			end if;
						
		when wr_avg   =>				NS <= init;	
			get_avg		<= '1';
			rst_pause	<= '1';
			
			if(chan_cnt < "100")then 	NS <= do_div;
			else                   	    NS <= avg_rdy;
			end if;
			
			inc_chan_cnt<= '1'; -- count increments when leaving this state

		when avg_rdy =>					NS <= clr_sum;
			rst_chan_cnt	<= '1';
			all_avgs_rdy	<= '1';
		-----------------------------------------------
		when others => 					NS <= init;
	end case;
end process;

----------------------------------------------------------------------------
-- MULTIPLEX INPUT CHANNELS INTO THE MATH FUNCTIONS 
NEW_VALUE_MUX: 
process(chan_cnt, CH1_IN, CH2_IN, CH3_IN, CH4_IN, 
		CH1_SUM, CH2_SUM, CH3_SUM, CH4_SUM) 
begin
	case chan_cnt is
		when "000" =>
			NEW_VALUE 	<= X"0000" & CH1_IN;
			CURRENT_SUM	<= CH1_SUM;
		when "001" =>
			NEW_VALUE 	<= X"0000" & CH2_IN;
			CURRENT_SUM	<= CH2_SUM;
		when "010" =>
			NEW_VALUE 	<= X"0000" & CH3_IN;
			CURRENT_SUM	<= CH3_SUM;
		when "011" =>
			NEW_VALUE 	<= X"0000" & CH4_IN;
			CURRENT_SUM	<= CH4_SUM;
		when others =>
			NEW_VALUE 	<= X"0000" & CH1_IN;
			CURRENT_SUM	<= CH1_SUM;
	end case;
end process;

-- MATH FUNCTIONS ----------------------------------------
ADD_VALUE_TO_SUM: unsigned_add32 
	port map(	dataa  => NEW_VALUE, 
				datab  => CURRENT_SUM, 
				result => Sum_Result );

DIVIDE_BY_AVG_LENGTH: unsigned_divide
	port map(	denom		=> 	Avg_Length,
				numer		=>	CURRENT_SUM,
				quotient	=>  Average_Result,
				remain		=>	REMAINDER );					                                                                                         	
----------------------------------------------------------
SUM_BUFFERS: 
process(clk, nreset, clr_sum_reg, get_sum, chan_cnt)
begin
	if(nreset='0')			then
		CH1_SUM		<= (others => '0');	
		CH2_SUM		<= (others => '0');			
		CH3_SUM		<= (others => '0');			
		CH4_SUM		<= (others => '0');			
		 
	elsif(rising_edge(clk))	then
		if   (clr_sum_reg = '1') then
			CH1_SUM		<= (others => '0');	
			CH2_SUM		<= (others => '0');			
			CH3_SUM		<= (others => '0');			
			CH4_SUM		<= (others => '0');			
		elsif(get_sum = '1') then
			case chan_cnt is
				when "000"  =>	CH1_SUM 	<= Sum_Result;
				when "001"  =>	CH2_SUM 	<= Sum_Result;
				when "010"  =>	CH3_SUM 	<= Sum_Result;
				when "011"  =>	CH4_SUM 	<= Sum_Result;
				when others =>	CH1_SUM 	<= Sum_Result;
			end case;
		end if;			   
	end if;
end process;

----------------------------------------------------------
AVG_BUFFERS: 
process(clk, nreset, get_avg, chan_cnt)
begin
	if(nreset='0')			then
		AVG1_BUF	<= (others => '0');	
		AVG2_BUF	<= (others => '0');			
		AVG3_BUF	<= (others => '0');			
		AVG4_BUF	<= (others => '0');			
		 
	elsif(rising_edge(clk))	then
		if(get_avg = '1') then
			case chan_cnt is
				when "000"  =>	AVG1_BUF 	<= Average_Result(15 downto 0);
				when "001"  =>	AVG2_BUF 	<= Average_Result(15 downto 0);
				when "010"  =>	AVG3_BUF 	<= Average_Result(15 downto 0);
				when "011"  =>	AVG4_BUF 	<= Average_Result(15 downto 0);
				when others =>	AVG1_BUF 	<= Average_Result(15 downto 0);
			end case;
		end if;			   
	end if;
end process;

-- OUTPUTS --------------------------------------------------------
OUTPUT_BUFFERS: 
process(clk, nreset)
begin
	if(nreset='0')			then
		AVG1_OUT	<= (others => '0');	
		AVG2_OUT	<= (others => '0');			
		AVG3_OUT	<= (others => '0');			
		AVG4_OUT	<= (others => '0');			
		
	elsif rising_edge(clk) 	then
		AVG_OUT_STRB 	<= all_avgs_rdy;
		AVG1_OUT  		<= AVG1_BUF;
		AVG2_OUT  		<= AVG2_BUF;
		AVG3_OUT  		<= AVG3_BUF;
		AVG4_OUT  		<= AVG4_BUF;                      
	end if;
end process;

--------------------------------------------
READ_SELECT_LED:
process(clk, nreset, RD_Avg_Reg, rd_sync)
begin
	if   (nreset='0')	then
		rd_sync <= (others =>'0');
	elsif rising_edge(clk)	then
		rd_sync 	<= RD_Avg_Reg;		
		RD_Avg_Strb <= RD_Avg_Reg and not rd_sync; --led
	end if;
end process;

--------------------------------------------     
LATCH_AVERAGE_TO_VME:       
process(clk, nreset, RD_Avg_Strb)
begin
	if   (nreset='0')	then 
		AVG_TO_VME	<= (others => '0');	
		
	elsif rising_edge(clk) 	then 		
		if   (RD_Avg_Strb(1) = '1') then AVG_TO_VME <= AVG1_BUF;
		elsif(RD_Avg_Strb(2) = '1') then AVG_TO_VME <= AVG2_BUF;
		elsif(RD_Avg_Strb(3) = '1') then AVG_TO_VME <= AVG3_BUF;
		elsif(RD_Avg_Strb(4) = '1') then AVG_TO_VME <= AVG4_BUF;
		end if;
	end if;
end process;	
	
end;             

