Library IEEE;
USE IEEE.std_logic_1164.all;
USE IEEE.std_logic_unsigned.all;

-- My Definitions --------------------------------
USE work.pkg_081006.all; 

------------------------------------------------------
-- MODULE DESCRIPTION
-- This module controls the presentation of test data
-- in place of the integrator ADC data for the purpose
-- of overall design verification.
--
-- A ROM memory is established that contains the test words.
-- The memory is setup using a *.mif file pointed to by the 
-- Altera Memory LPM Megafunction.
--
-- The memory is organized into 4 sets of 512 each 16 bit words
-- The input bits Test_Sel(1..0) select which of the 4 sets
-- are used for the current test. A counter is implemented
-- as the memory pointer and advances with occurance of Next_Word
-- (aka ADC_DONE).  The memory pointer is reset when Test_Done
-- (aka FIFO_FULL) goes active.

--==========================================================


------------------------------------------------------

entity Test_Vector_Memory is 
	port( 
		clk				: in std_logic;
		nreset			: in std_logic;
		
		Test_Sel		: in std_logic_vector(1 downto 0);	
		Run_DAQ			: in std_logic;			
		Next_Word		: in std_logic; -- step to next test word
		Test_Done		: in std_logic; -- end of record, reset to first word
		
		Test1_out		: out std_logic_vector(15 downto 0);
		Test2_out		: out std_logic_vector(15 downto 0);
		Test3_out		: out std_logic_vector(15 downto 0);
		Test4_out		: out std_logic_vector(15 downto 0)
		);
	end Test_Vector_Memory ;
	
--########################################################
--========================================================
--				A R C H I T E C T U R E
--========================================================
--########################################################
architecture main of Test_Vector_Memory  is

-- Strobe Sync Signals ------------------------------------
	signal mem_ptr  	: std_logic_vector(8 downto 0);
	signal mem_ptr_inc	: std_logic;
	signal mem_ptr_rst	: std_logic;
	
	signal mem_data_out : std_logic_vector(15 downto 0);

-- Syncronized Input Signals -------------------------------
	signal start_tst	: std_logic;
	signal nxt_wrd		: std_logic;
	signal tst_done		: std_logic;	
------------------------------------------------------------
-- ALTERA COMPONENTS ---------------------------------------
-- The following component is the RAM memory
-- It was created using the LPM Megafunction Wizard.
-- The following VHDL file was generated and must be included 
-- in the project.
--                  Test_Vector_ROM.vhd
--
-- The Megafunction Wizard also generate *.cmp files for each
-- from which I cut and pasted the component declarations below.
----------------------------------------------------------------
	component Test_Vector_ROM_FE
		PORT
		(
			address		: IN STD_LOGIC_VECTOR (10 DOWNTO 0);
			clock		: IN STD_LOGIC ;
			q			: OUT STD_LOGIC_VECTOR (15 DOWNTO 0)
		);
	end component;

-- SETUP_STATE_MACHINE ---------------------------------------
	type STATE_TYPE is
	(idle, inc_ptr, wait0, done);
	signal CS, NS : STATE_TYPE;
	
--########################################################
--========================================================
--						B E G I N
--========================================================
--########################################################

begin

--**-- --**--   --**--   --**--   --**--   --**--   --**-- 
-- STROBE SYNCHRONIZATION --------------------------------

SYNC_START: led -- syncronize the Transf_Done.
port map(	sig_in  => Run_DAQ, 
			clk     => clk, 
			clrn    => '1', 
			pls_out => start_tst); 
 
SYNC_NEXT: led -- syncronize the signal to start setup sequence.
port map(	sig_in  => Next_Word, 
			clk     => clk, 
			clrn    => '1', 
			pls_out => nxt_wrd); 
			
SYNC_DONE: led -- syncronize the Transf_Done.
port map(	sig_in  => not Run_DAQ OR Test_Done, 
			clk     => clk, 
			clrn    => '1', 
			pls_out => tst_done); 
			
MEMORY_POINTER: 
process(clk, nreset, mem_ptr_inc, mem_ptr_rst)
begin
	if(nreset = '0')			 then mem_ptr <= (others => '0');
	elsif(rising_edge(clk))		 then 
		if	 (mem_ptr_rst = '1') then mem_ptr <= (others => '0');
		elsif(mem_ptr_inc = '1') then mem_ptr <= mem_ptr + 1;
		else						  mem_ptr <= mem_ptr;
		end if;
	end if;
end process;

--**--   --**--   --**--   --**--   --**--   --**--   --**--   
-- SETUP RAM MEMORY ----------------------------------------
SETUP_ROM: Test_Vector_ROM_FE -- ROM memory that holds test words
port map(	address  => Test_Sel & mem_ptr,
			clock	 => clk,
			q		 => mem_data_out);        
  
--########################################################
--========================================================
--					STATE MACHINE
--========================================================
--########################################################

SM_REG: process(clk, nreset)
begin
	if(nreset='0')				then CS <= idle;
	elsif(rising_edge(clk))	then CS <= NS;
	end if;
end process;

SM_NEXT_STATE: 
process(CS, start_tst, nxt_wrd, tst_done)
begin
	mem_ptr_rst <= '0'; -- default value
	mem_ptr_inc <= '0'; -- default value
	
	case CS is
		when idle =>
			mem_ptr_rst <= '1'; -- pointer held at zero when idle
			if(start_tst = '1')then 	NS <= inc_ptr;
			else                   	    NS <= idle;
			end if;
			
		when inc_ptr => 				NS <= wait0;
			mem_ptr_inc <= '1'; -- increment the pointer

		when wait0 => 				
			if   (nxt_wrd  = '1') then 	NS <= inc_ptr;
			elsif(tst_done = '1') then	NS <= done;
			else                   	    NS <= wait0;
			end if;

		when done => 					NS <= idle;

		when others => 					NS <= idle;
	end case;
end process;

SYNCRONIZED_OUTPUTS: process(clk)
begin
	if(rising_edge(clk)) then
		Test1_out  <= mem_data_out(15 downto 0); 
		Test2_out  <= mem_data_out(14 downto 0) & '0';  
		Test3_out  <= mem_data_out(13 downto 0) & "00";
		Test4_out  <= mem_data_out(12 downto 0) & "000";
	end if;
end process;

end main;		
		
			